#include	<stdio.h>
#include	<string.h>
#include	<stdlib.h>
#include	<ctype.h>
#include	<math.h>

#include	<tsdb.h>
#include	<ace/lisp.h>
#include	<assert.h>

#include	"fill.h"
#include	"tokens.h"

int	gold_items[] = {10003,10023,10027,10038,10043,10046,10050,10051,10053,10059,10060,10061,10086,10121,10122,10125,10128,10131,10139,10143,10145,10149,10164,10167,10171,10183,10190,10214,10218,10221,10223,10225,10230,10246,10250,10262,10282,10287,10303,10320,10328,10337,10340,11012,11014,11024,11028,11029,11030,11033,11035,11037,11045,11046,11076,11099,11108,11121,11128,11132,11133,11146,11147,11154,11155,11158,11169,11172,11187,11189,11192,11193,11195,11211,11223,11224,11227,11232,11257,11264,11266,11269,11270,11276,11277,11345,11365,11392,11393,11403,11411,11414,11416,11418,11420,11423};
int	ngold_items = sizeof(gold_items) / sizeof(int);
int is_gold_item(char	*id)
{
	int	iid = atoi(id);
	int j;
	for(j=0;j<ngold_items;j++)
		if(iid == gold_items[j])return 1;
	return 0;
}

int	dcmp(const double	*a, const double	*b)
{
	if(*a > *b)return -1;
	if(*a < *b)return 1;
	return 0;
}

double	confidence_threshold = 0.0, token_inherit_threshold = 0.0;

int	predicted_tokens = 0;

unmark_tokens(struct starsem_token_list	*toks)
{
	int i;
	for(i=0;i<toks->count;i++)
	{
		toks->tokens[i].in_gold_scope = 0;
		toks->tokens[i].in_auto_scope = 0;
		toks->tokens[i].in_w_scope = 0;
		toks->tokens[i].is_cue = 0;
		toks->tokens[i].p_in = 0;
		toks->tokens[i].p_out = 0;
		toks->tokens[i].inherit = 0;
	}
}

wunmark_tokens(struct starsem_token_list	*toks)
{
	int i;
	for(i=0;i<toks->count;i++)
		toks->tokens[i].in_w_scope = 0;
}

conjoin_markings(double	p, struct starsem_token_list	*toks)
{
	int i;
	for(i=0;i<toks->count;i++)
	{
		if(toks->tokens[i].in_w_scope)
			toks->tokens[i].p_in += p;
		else toks->tokens[i].p_out += p;
	}
}

final_vote_markings(struct starsem_token_list	*toks, double	thresh)
{
	int i;
	for(i=0;i<toks->count;i++)
	{
		if(toks->tokens[i].p_out < thresh && toks->tokens[i].p_in < thresh)
		{
			toks->tokens[i].inherit = 1;
			//printf("token %s should inherit [%f vs %f]\n", toks->tokens[i].text, toks->tokens[i].p_in, toks->tokens[i].p_out);
		}
		else if(toks->tokens[i].p_out >= toks->tokens[i].p_in)
			toks->tokens[i].in_w_scope = 0;
		else toks->tokens[i].in_w_scope = 1;
	}
}

void mark_token(struct starsem_token_list	*toks, int	left, int	right, int	mark)
{
	int i;
	for(i=0;i<toks->count;i++)
		if(toks->tokens[i].left >= left && toks->tokens[i].right <= right)
		{
			if(mark==1)toks->tokens[i].in_gold_scope = 1;
			if(mark==2)toks->tokens[i].in_auto_scope = 1;
			if(mark==3)toks->tokens[i].in_w_scope = 1;
			if(mark==-1) { toks->tokens[i].in_w_scope = 0; toks->tokens[i].is_cue = 1; }
		}
}

void mark_cue(struct starsem_token_list	*toks, int	left, int	right)
{
	int i;
	for(i=0;i<toks->count;i++)
		if(toks->tokens[i].left <= left && toks->tokens[i].right >= right)
		{
			toks->tokens[i].in_w_scope = 0;
			toks->tokens[i].is_cue = 1;
		}
}

int count_marks(struct starsem_token_list	*toks, int mark)
{
	int i, n = 0;
	for(i=0;i<toks->count;i++)
	{
		if(mark==1 && toks->tokens[i].in_gold_scope)n++;
		if(mark==2 && toks->tokens[i].in_auto_scope)n++;
		if(mark==3 && toks->tokens[i].in_w_scope)n++;
	}
	return n;
}

int is_punct_token(struct starsem_token	*t)
{
	char	*c = t->text;
	while(*c && ispunct(*c))c++;
	if(*c)return 0;
	return 1;
}

tweak_markings(struct starsem_token_list	*toks)
{
	int i, j;
	// unmark punctuation tokens preceding out-of-scope tokens
	for(i=0;i<=toks->count;i++)
	{
		if(i==toks->count || !toks->tokens[i].in_w_scope)
		{
			for(j=i-1;j>=0 && toks->tokens[j].in_w_scope && is_punct_token(&toks->tokens[j]);j--)
				toks->tokens[j].in_w_scope = 0;
		}
	}
	// unmark punctuation tokens following out-of-scope tokens
	for(i=-1;i<toks->count;i++)
	{
		if(i==-1 || !toks->tokens[i].in_w_scope)
		{
			for(j=i+1;j<toks->count && toks->tokens[j].in_w_scope && is_punct_token(&toks->tokens[j]);j++)
				toks->tokens[j].in_w_scope = 0;
		}
	}
	// unmark "n't" tokens
	for(i=0;i<toks->count;i++)
	{
		if(toks->tokens[i].in_w_scope && !strcasecmp(toks->tokens[i].text, "n't"))
			toks->tokens[i].in_w_scope = 0;
	}
}

find_given_scopes(struct s_exp	*nscopes, char	*id, struct starsem_token_list	*toks)
{
	//printf("scope of cue %s:\n", id);
	assert(nscopes->type == s_exp_list);
	int i;
	for(i=1;i<nscopes->len;i++)
	{
		struct s_exp	*I = nscopes->list[i];
		assert(I->type == s_exp_list & I->len == 2);
		struct s_exp	*Iid = I->list[0];
		struct s_exp	*Isp = I->list[1];
		assert(Iid->type == s_exp_cons && Isp->type == s_exp_list);
		assert(!strcmp(Iid->list[0]->value, ":id"));
		if(strcmp(Iid->list[1]->value, id))continue;
		assert(Isp->len > 1);
		assert(Isp->list[0]->type == s_exp_id);
		assert(!strcmp(Isp->list[0]->value, ":span"));
		int j;
		for(j=1;j<Isp->len;j++)
		{
			struct s_exp	*sp = Isp->list[j];
			assert(sp->type == s_exp_list && sp->len == 3);
			assert(sp->list[0]->type == s_exp_id);
			assert(sp->list[1]->type == s_exp_id);
			assert(sp->list[2]->type == s_exp_id);
			int from = atoi(sp->list[0]->value);
			int to = atoi(sp->list[1]->value);
			char	*type = sp->list[2]->value;
			//printf("scope: %d - %d [%s]\n", from, to, type);
			mark_token(toks, from, to, 1);
		}
	}
}

struct s_exp	*get_key(struct s_exp	*alist, char	*key)
{
	int i;
	assert(alist->type == s_exp_list);
	for(i=0;i<alist->len;i++)
	{
		struct s_exp	*s = alist->list[i];
		assert(s->type == s_exp_cons || (s->type == s_exp_list && s->len > 1));
		struct s_exp	*sk = s->list[0];
		struct s_exp	*sv = s->list[1];
		assert(sk->type == s_exp_id);
		if(!strcasecmp(sk->value, key))
			return sv;
	}
	fprintf(stderr, "no such key '%s'\n", key);
	exit(-1);
}

char	*get_key_str(struct s_exp	*alist, char	*key)
{
	struct s_exp	*s = get_key(alist, key);
	assert(s != NULL);
	return s->value;
}

struct starsem_token_list	*extract_tokens(struct s_exp	*toklist)
{
	struct starsem_token_list	*sstl = calloc(sizeof(*sstl),1);
	assert(toklist->type == s_exp_list);
	int i;
	sstl->count = toklist->len;
	sstl->tokens = calloc(sizeof(struct starsem_token),sstl->count);
	for(i=0;i<toklist->len;i++)
	{
		struct starsem_token	*sst = &sstl->tokens[i];
		struct s_exp	*tok = toklist->list[i];
		char	*form = get_key_str(tok, ":form");
		int	start = atoi(get_key_str(tok, ":start"));
		int	end = atoi(get_key_str(tok, ":end"));
		sst->left = start;
		sst->right = end;
		sst->text = strdup(form);
		sst->in_gold_scope = 0;
		sst->in_auto_scope = 0;
		//printf("[%d - %d]	%s\n", start, end, form);
	}
	return sstl;
}

/*
print_sentence(struct starsem_token_list	*tokens)
{
	int i;
	for(i=0;i<tokens->count;i++)
	{
		struct starsem_token	*T = &tokens->tokens[i];
		printf("[%5d - %5d / %c%c%c ]	%s\n", T->left, T->right, T->in_gold_scope?'G':' ', T->in_auto_scope?'A':' ', T->in_w_scope?'W':' ', T->text);
	}
}*/

/*
output_connl(struct starsem_token_list	*tokens, char		*sid)
{
	FILE	*cf = fopen("/tmp/connl.txt", "a");
	int i;
	for(i=0;i<tokens->count;i++)
	{
		struct starsem_token	*T = &tokens->tokens[i];
		fprintf(cf, "%s	%d	%s	%s	_	_", sid, i, T->text, T->text);
		if(T->is_cue)fprintf(cf, "	%s", T->text);
		else fprintf(cf, "	_");
		if(T->in_w_scope)fprintf(cf, "	%s", T->text);
		else fprintf(cf, "	_");
		fprintf(cf, "	_\n");
	}
	fprintf(cf, "\n");
	fclose(cf);
}
*/

void output_connl(struct starsem_token_list	*tokens, char	*sid)
{
	int	nw=0, ni = 0, nc = 0;
	int i;
	for(i=0;i<tokens->count;i++)
	{
		if(tokens->tokens[i].in_w_scope)nw++;
		if(tokens->tokens[i].inherit)ni++;
		if(tokens->tokens[i].is_cue)nc++;
	}
	if(!nw)return;
	if(ni == tokens->count)return;
	if(tokens->count == 0)return;
	if(nc == 0)return;
	char	pathname[1024];
	sprintf(pathname, "connl.txt");
	FILE	*cf = fopen(pathname, "a");
	if(!cf) { perror(pathname); exit(-1); }
	for(i=0;i<tokens->count;i++)
	{
		struct starsem_token	*T = &tokens->tokens[i];
		fprintf(cf, "%s	%d	%s	%s	%d\n", sid, i, T->text, T->inherit?"I":(T->in_w_scope?"1":"0"), T->is_cue?1:0);
		if(!T->inherit)predicted_tokens++;
	}
	fprintf(cf, "\n");
	fclose(cf);
}

printcn(char	*c, int	n)
{
	while(n-->0)printf("%s",c);
}

print_sentence(struct starsem_token_list	*tokens)
{
	int	MAX_WIDTH=65;
	int i, width = 0;
	int printed = 0, start = 0;
	while(printed < tokens->count)
	{
		width = 0;
		while(width < MAX_WIDTH && printed < tokens->count)
		{
			width += 1 + strlen(tokens->tokens[printed].text);
			printed++;
		}
		for(i=start;i<printed;i++)printf(" %s", tokens->tokens[i].text);
		printf("\n");
		for(i=start;i<printed;i++){printf(" ");printcn(tokens->tokens[i].in_gold_scope?"G":"-", strlen(tokens->tokens[i].text));}
		printf("\n");
		for(i=start;i<printed;i++){printf(" ");printcn(tokens->tokens[i].in_w_scope?"W":"-", strlen(tokens->tokens[i].text));}
		printf("\n");
		printf("\n");
		start = printed;
	}
}

int	fp1, tp1, fn1;
int	fp2, tp2, fn2;
int	ns1, ns2;

int	score_sentence(struct starsem_token_list	*tokens)
{
	int	nauto=0, nw=0;
	int	nerrors = 0;
	int i;
	for(i=0;i<tokens->count;i++)
	{
		if(tokens->tokens[i].in_auto_scope)nauto++;
		if(tokens->tokens[i].in_w_scope)nw++;
	}
	if(nauto)ns1++;
	if(nw)ns2++;
	for(i=0;i<tokens->count;i++)
	{
		struct starsem_token	*T = &tokens->tokens[i];
		if(nauto)
		{
			if(T->in_gold_scope && T->in_auto_scope)tp1++;
			if(!T->in_gold_scope && T->in_auto_scope)fp1++;
			if(T->in_gold_scope && !T->in_auto_scope)fn1++;
		}
		if(nw)
		{
			if(T->in_gold_scope && T->in_w_scope)tp2++;
			if(!T->in_gold_scope && T->in_w_scope)fp2++;
			if(T->in_gold_scope && !T->in_w_scope)fn2++;
			if(T->in_gold_scope != T->in_w_scope)nerrors++;
		}
	}
	return nerrors;
}

show_pr(int tp, int fp, int fn)
{
	double	P, R, F1;
	P = (double)tp / (tp+fp);
	R = (double)tp / (tp+fn);
	F1 = (P && R)?(2.0 / (1.0 / P + 1.0 / R)):0;
	printf("P = %.2f%% ; R = %.2f%% ; F1 = %.2f%%\n", 100*P, 100*R, 100*F1);
}

report_scores()
{
	printf("method 1: picking a scoped-mrs\n");
	printf("considering just the %d items where this method claimed at least one token was in scope...\n", ns1);
	show_pr(tp1, fp1, fn1);
	printf("method 2: traversing underspecified-mrs\n");
	printf("considering just the %d items where this method claimed at least one token was in scope...\n", ns2);
	show_pr(tp2, fp2, fn2);

	printf("made connl predictions about %d tokens\n", predicted_tokens);
}

main(int	argc, char	*argv[])
{
	char	pathname[1024];
	sprintf(pathname, "echo > connl.txt");
	system(pathname);
	ace_load_grammar("erg-1212.dat");
extern char	mrs_line_breaks, mrs_tab;
	mrs_line_breaks = '\n'; mrs_tab = '\t';
	assert(argc>=2);
	if(argc>=3)confidence_threshold = atof(argv[2]);
	if(argc>=4)token_inherit_threshold = atof(argv[3]);
	if(argc > 4)
	{
		int k;
		for(k=4;k<argc;k++)enable_fc_rel(argv[k]);
	}
	else
	{
		enable_fc_rel("_whether_x_rel/ARG1");
		enable_fc_rel("_in+order+to_x_rel/ARG1");
		enable_fc_rel("_when_x_subord_rel/ARG1");
		enable_fc_rel("_because_x_rel/ARG1");
		enable_fc_rel("_with_x_subord_rel/ARG1");
		enable_fc_rel("_although_x_rel/ARG1");
		enable_fc_rel("_unless_x_rel/ARG1");
		enable_fc_rel("_until_x_h_rel/ARG1");
		enable_fc_rel("_as_x_subord_rel/ARG1");
		enable_fc_rel("appos_rel/ARG1");
	}
	struct tsdb	*T = load_tsdb(argv[1]);
	struct relation	*item = get_relation(T, "item");
	int	i_id = get_field(item, "i-id", "integer");
	int	i_comment = get_field(item, "i-comment", "string");
	int	i_tokens = get_field(item, "i-tokens", "string");
	struct relation	*parse = get_relation(T, "parse");
	int	p_i_id = get_field(parse, "i-id", "integer");
	int p_p_id = get_field(parse, "parse-id", "integer");
	struct relation	*result = get_relation(T, "result");
	int	r_id = get_field(result, "result-id", "integer");
	int	r_p_id = get_field(result, "parse-id", "integer");
	int r_deriv = get_field(result, "derivation", "string");
	int r_mrs = get_field(result, "mrs", "string");
	int r_flags = get_field(result, "flags", "string");
	int i;
	for(i=0;i<item->ntuples;i++)
	{
		char	*id = item->tuples[i][i_id];

		// XXX hack
		//if(!is_gold_item(id))continue;

		// XXX hack
		//if(atoi(id) >= 11000)continue;
		char	*i_tokens_str = item->tuples[i][i_tokens];
		struct s_exp *lisp_toklist = parse_lisp(&i_tokens_str, 1);
		char	*comment = item->tuples[i][i_comment];
		if(!*comment)
		{
			printf("no negation annotation for %s\n\n", id);
			continue;
		}
		char	**ptup = tsdb_lookup_relation_with_key(parse, p_i_id, id);
		if(!ptup)
		{
			fprintf(stderr, "no parse tuple for iid %s\n", id);
			continue;
		}
		assert(ptup);
		char	*pid = ptup[p_p_id];
		printf("--- ITEM %s ---\n", id);
		struct starsem_token_list	*toklist = extract_tokens(lisp_toklist);
		struct s_exp	*s = parse_lisp(&comment, 1);
		assert(s != NULL);
		assert(s->type = s_exp_list && s->len == 3);
		struct s_exp	*s_nscopes = s->list[1];
		struct s_exp	*s_ncues = s->list[2];
		assert(s_nscopes->type == s_exp_list && s->len > 0);
		assert(s_ncues->type == s_exp_list && s->len > 0);
		assert(s_nscopes->list[0]->type == s_exp_id);
		assert(!strcasecmp(s_nscopes->list[0]->value, ":nscopes"));
		assert(s_ncues->list[0]->type == s_exp_id);
		assert(!strcasecmp(s_ncues->list[0]->value, ":ncues"));
		if(s->len == 1)continue;


		int j;
		for(j=1;j<s_ncues->len;j++)
		{
			// for each cue...

			struct s_exp	*s_ncue = s_ncues->list[j];
			assert(s_ncue->type == s_exp_list);
			assert(s_ncue->len == 2);
			struct s_exp	*id_cons = s_ncue->list[0];
			assert(id_cons->type == s_exp_cons);
			assert(id_cons->list[0]->type == s_exp_id);
			assert(id_cons->list[1]->type == s_exp_id);
			assert(!strcasecmp(id_cons->list[0]->value, ":id"));
			char	*cue_id = id_cons->list[1]->value;

			struct s_exp	*s_spans = s_ncue->list[1];
			assert(s_spans->type == s_exp_list);
			printf("cue %s has %d spans\n", cue_id, s_spans->len - 1);
			unmark_tokens(toklist);
			find_given_scopes(s_nscopes, cue_id, toklist);

			// get the top 10 probabilities
			double	p_distribution[10];
			double	p_best = 0;
			char	*id_best = NULL;
			int		nreadings = 0;
			void visit_result_p(char	**rtup)
			{
				int have_confidence = (rtup[r_flags] && *rtup[r_flags])?1:0;
				if(have_confidence && nreadings<10)
				{
					double	p = atof(rtup[r_flags]);
					p_distribution[nreadings++] = p;
					//printf("   P = %f\n", p);
					if(p >= p_best) { p_best = p; id_best = rtup[r_id]; }
				}
			}
			tsdb_visit_relation_with_key(result, r_p_id, pid, visit_result_p);
			//printf("got %d probabilities\n", nreadings);
			qsort(p_distribution, nreadings, sizeof(double), dcmp);
			double	p_ratio = 1.0, entropy = 0;
			if(nreadings > 1)
			{
				if(p_distribution[1] > 0)
					p_ratio = p_distribution[0] / p_distribution[1];
				else p_ratio = 100000;
				int k;
				double	p_norm = 0;
				for(k=0;k<nreadings;k++)p_norm += p_distribution[k];
				if(p_norm > 0)
				{
					for(k=0;k<nreadings;k++)
						entropy += - (p_distribution[k] / p_norm) * log2(p_distribution[k] / p_norm);
				}
			}
			else p_ratio = 100000, entropy = 0;

			int	only_use_first_reading = 0;
			if(nreadings > 0)
			{
				//printf("top P   = %f\n", p_distribution[0]);
				//printf("P ratio = %f\n", p_ratio);
				//printf("entropy = %f\n", entropy);
				double	entropy2 = entropy / log2(nreadings);
				//printf("entropy / log_2(%d) = %f\n", nreadings, entropy2);

				if(p_distribution[0] < confidence_threshold)
				{
					//printf("not confident enough.\n");
					continue;
				}
				only_use_first_reading = 1;
				assert(id_best);
			}

			void visit_result(char	**rtup)
			{
				// for each reading of the sentence...
				char	*sderiv = NULL, *smrs = NULL;
				struct tree	*deriv = NULL;
				struct mrs	*mrs = NULL;
				int	have_confidence = 0;
				double	confidence = 1;

				if(!rtup)return;
				clear_slab();
				clear_mrs();
				if(only_use_first_reading)
				{
					if(!strcmp(id_best, rtup[r_id]))confidence = confidence_threshold + 1;
					else confidence = 0;
				}
				else
				{
					have_confidence = (rtup[r_flags] && *rtup[r_flags])?1:0;
					confidence = have_confidence?atof(rtup[r_flags]):1.0;
				}
				if(confidence >= confidence_threshold)
				{
					sderiv = strdup(rtup[r_deriv]);
					deriv = string_to_tree(sderiv);
					free(sderiv);
					smrs = strdup(rtup[r_mrs]);
					mrs = parse_mrs(smrs);
					free(smrs);
				}
				else return;
				wunmark_tokens(toklist);
				int k;
				// crawl from each span of the cue
				for(k=1;k<s_spans->len;k++)
				{
					struct s_exp	*s_span = s_spans->list[k];
					assert(s_span->type == s_exp_list && s_span->len == 3);
					assert(s_span->list[0]->type == s_exp_id);
					assert(s_span->list[1]->type == s_exp_id);
					assert(s_span->list[2]->type == s_exp_id);
					char	*from = s_span->list[0]->value;
					char	*to = s_span->list[1]->value;
					char	*type = s_span->list[2]->value;
					printf(" span %d: from %s to %s is a cue of type %s\n", k, from, to, type);
					fflush(stdout);

					alt_find_scope(mrs, deriv, type, atoi(from), atoi(to), toklist);
				}
				// remove each span of the cue from the scope
				for(k=1;k<s_spans->len;k++)
				{
					struct s_exp	*s_span = s_spans->list[k];
					char	*from = s_span->list[0]->value;
					char	*to = s_span->list[1]->value;
					mark_cue(toklist, atoi(from), atoi(to));
				}
				tweak_markings(toklist);
				conjoin_markings(confidence, toklist);
			}
			//char	**rtup = tsdb_lookup_relation_with_key(result, r_p_id, pid);
			tsdb_visit_relation_with_key(result, r_p_id, pid, visit_result);
			final_vote_markings(toklist, token_inherit_threshold);
			int errors = score_sentence(toklist);
			output_connl(toklist, id);
			printf(" (by the way, this was ITEM %s )\n", id);
			if(errors)printf("%d errors\n", errors);
			print_sentence(toklist);
		}
		printf("\n\n");
	}
	report_scores();
}
