#include	<stdlib.h>
#include	<stdio.h>
#include	<string.h>
#include	<assert.h>

#include	<ace/mrs.h>
#include	<ace/lisp.h>
#include	<ace/reconstruct.h>
#include	<ace/conf.h>

#include	"tokens.h"
#include	"fill.h"

char	*mrs_find_var_prop(struct mrs_var	*v, char	*prop)
{
	int i;
	for(i=0;i<v->nprops;i++)
		if(!strcasecmp(v->props[i].name, prop))
			return v->props[i].value;
	return NULL;
}

struct mrs_var	*traverse_qeq(struct mrs	*m, struct mrs_var	*hi)
{
	int i;
	for(i=0;i<m->nhcons;i++)
		if(m->hcons[i].hi == hi)
			return m->hcons[i].lo;
	return NULL;
}

int	acceptable_pred(char	*pred)
{
	if(strstr(pred, "_d_"))return 0;
	return 1;
}

int	nfcpreds;
char	**fcpreds;
char	**fcargs;
enable_fc_rel(char	*predarg)
{
	//if(!fcpreds)fcpreds = hash_new("functor-crawl-preds");
	//hash_add(fcpreds, strdup(pred), (void*)0x1);
	char	pred[1024], arg[256];
	assert(2 == sscanf(predarg, "%[^/]/%[^ ]", pred, arg));
	nfcpreds++;
	fcpreds = realloc(fcpreds, sizeof(char*)*nfcpreds);
	fcpreds[nfcpreds-1] = strdup(pred);
	fcargs = realloc(fcargs, sizeof(char*)*nfcpreds);
	fcargs[nfcpreds-1] = strdup(arg);
}

int	acceptable_fcpred(char	*pred, char	*arg)
{
	if(strstr(pred, "_modal"))return 1;
	if(strstr(pred, "_why_a_1_rel"))return 1;
	int i;
	//char	query[1024];
	//sprintf(query, "%s/%s", pred, arg);
	for(i=0;i<nfcpreds;i++)
		if(!strcasecmp(arg, fcargs[i]) && strstr(pred, fcpreds[i]))return 1;
	//if(fcpreds && hash_find(fcpreds, pred))return 1;
	/*if(strstr(pred, "_with_x_subord_rel"))return 1;
	if(strstr(pred, "although_x_rel"))return 1;
	if(strstr(pred, "_until_x_h_rel"))return 1;*/
	return 0;
}

int	unacceptable_label_sharer(struct mrs_ep	*candidate, struct mrs_ep	*neg_rel)
{
	// be careful about EPs that share their label with the negation cue
	if(candidate->lbl != neg_rel->lbl)return 0;
	int i, j;
	// if one is an argument of the other, good.
	for(i=0;i<candidate->nargs;i++)
		for(j=0;j<neg_rel->nargs;j++)
			if(candidate->args[i].value == neg_rel->args[j].value
			&& (!strcmp(candidate->args[i].name, "ARG0") || !strcmp(neg_rel->args[j].name, "ARG0"))
			//&& (!strcmp(candidate->args[i].name, "ARG1") || !strcmp(neg_rel->args[j].name, "ARG1"))
			)
				return 0;
	// no good.
	return 1;
}

int unacceptable_quantifier(struct mrs	*m, struct mrs_ep	*candidate, struct mrs_ep	*neg)
{
	if(strstr(candidate->pred, "_q_"))
	{
		struct mrs_var	*rstr = mrs_find_ep_role(candidate, "RSTR");
		assert(rstr);
		rstr = traverse_qeq(m, rstr)?:rstr;
		if(rstr == neg->lbl)
		{
			int i, j;
			// if one's ARG0 is the other's ARGn, OK
			for(i=0;i<candidate->nargs;i++)
				for(j=0;j<neg->nargs;j++)
					if(candidate->args[i].value == neg->args[j].value
					&& (!strcmp(candidate->args[i].name, "ARG0") || !strcmp(neg->args[j].name, "ARG0"))
					)
				return 0;
			// otherwise not OK
			return 1;
			/*struct mrs_var	*rarg0 = mrs_find_ep_role(candidate, "ARG0");
			struct mrs_var	*rarg1 = mrs_find_ep_role(candidate, "ARG1");
			struct mrs_var	*neg_rel_arg0 = mrs_find_ep_role(neg, "ARG0");
			struct mrs_var	*neg_rel_arg1 = mrs_find_ep_role(neg, "ARG1");
			if(rarg0 != neg_rel_arg1 && rarg0 != neg_rel_arg0)
				return 1;*/
		}
	}
	return 0;
}

int	mark_by_label(struct mrs	*m, struct mrs_var	*lbl, int	*enabled, struct mrs_ep	*neg)
{
	int i, n = 0;
	for(i=0;i<m->nrels;i++)
	{
		if(m->rels[i].lbl == lbl && !enabled[i] && acceptable_pred(m->rels[i].pred))
		{
			if(unacceptable_label_sharer(&m->rels[i], neg))continue;
			printf("enabled %s by label\n", m->rels[i].pred);
			enabled[i] = 1;
			n++;
		}
	}
	return n;
}

int	mark_by_arg0(struct mrs	*m, struct mrs_var	*arg0, int	*enabled, struct mrs_ep	*neg)
{
	int i, n = 0;
	for(i=0;i<m->nrels;i++)
	{
		struct mrs_var	*rarg0 = mrs_find_ep_role(&m->rels[i], "ARG0");
		if(unacceptable_quantifier(m, &m->rels[i], neg))continue;
		if(unacceptable_label_sharer(&m->rels[i], neg))continue;
		if(rarg0 == arg0 && !enabled[i] && acceptable_pred(m->rels[i].pred))
		{
			printf("enabled %s by arg0\n", m->rels[i].pred);
			enabled[i] = 1;
			n++;
		}
	}
	return n;
}

void disable_zero_pronouns(struct mrs	*m, int	*enabled)
{
	int i;
	for(i=0;i<m->nrels;i++)
	{
		struct mrs_ep	*e = &m->rels[i];
		struct mrs_var	*x = mrs_find_ep_role(e, "ARG0");
		if(!x)continue;
		char	*pt = mrs_find_var_prop(x, "PRONTYPE");
		if(!pt)continue;
		if(!strcasecmp(pt, "zero_pron"))enabled[i] = 0;
	}
}

int	span_in_scope(int	from, int	to, struct starsem_token_list	*toks)
{
	int i;
	for(i=0;i<toks->count;i++)
		if(toks->tokens[i].left >= from && toks->tokens[i].right <= to)
			if(toks->tokens[i].in_w_scope)
				return 1;
	return 0;
}

int vacuous_in_scope1(struct lexeme	*lex, struct tree	*v, struct starsem_token_list	*toks, char	*rule)
{
	extern char	*trim_string(char*);
	rule = trim_string(rule);
	if(*rule==0)return 0;
	if(*rule==';')return 0;
	char	*type = strtok(rule, " \t");
	struct type	*t = lookup_type(type);
	assert(t);
	if(!descendent_of(lex->lextype->tnum, t))return 0;
	//printf("checking %s against: %s\n", lex->word, rule);
	char	*cons;
	while(cons = strtok(NULL, " \t"))
	{
		struct tree	*n = v;
		//printf(" %s  ?\n", cons);
		while(*cons=='^')
		{
			cons++;
			n = n->parent;
			if(!n)return 0;
		}
		while(*cons=='.')
		{
			cons++;
			assert(isdigit(*cons));
			int d = (*cons) - '0';
			cons++;
			assert(d < n->ndaughters);
			n = n->daughters[d];
			assert(n->ndaughters > 0);
		}
		assert(*cons == ':');
		cons++;
		if(!strcmp(cons, "in-scope"))
		{
			if(n->edge_id != 1)return 0;// doesn't work yet, fixme.
			//struct tree	*lh = tree_lex_head(n);
			//if(!span_in_scope(lh->cfrom, lh->cto, toks))return 0;
		}
		else
		{
			//printf(" label is %s\n", n->label);
			if(strcmp(n->label, cons))return 0;
		}
	}
	return 1;
}

int vacuous_in_scope(struct lexeme	*lex, struct tree	*v, struct starsem_token_list	*toks)
{
	FILE	*f = fopen("vac-rules.txt", "r");
	if(!f) { perror("vac-rules.txt"); exit(-1); }
	char	rule[10240];
	while(fgets(rule, 10239, f))
		if(vacuous_in_scope1(lex, v, toks, rule))
		{
			fclose(f);
			return 1;
		}
	fclose(f);
	return 0;
}

int is_coord_rule(char	*name)
{
	// dan says the right thing is to look at the CTYPE hierarchy
	if(strstr(name, "crd"))return 1;
	else return 0;
}

int	decorate_deriv(struct tree	*t, struct starsem_token_list	*toks)
{
	int in_scope = 0;
	if(t->ndaughters > 1 || !(t->ndaughters==1 && t->daughters[0]->ndaughters == 0))
	{
		int	i;
		int is_coord = is_coord_rule(t->label);
		for(i=0;i<t->ndaughters;i++)
		{
			int s = decorate_deriv(t->daughters[i], toks);
			if(is_coord)in_scope |= s;
			else if(tree_lex_head(t) == tree_lex_head(t->daughters[i]))
				in_scope = s;
		}
	}
	else in_scope = span_in_scope(t->cfrom, t->cto, toks);
	t->edge_id = in_scope?1:0;
	return in_scope;
}

void mark_vacuous(struct mrs	*m, int	*enabled, struct tree	*deriv, struct starsem_token_list	*toks)
{
	if(deriv->ndaughters > 1 || !(deriv->ndaughters==1 && deriv->daughters[0]->ndaughters == 0))
	{
		int i;
		for(i=0;i<deriv->ndaughters;i++)
			mark_vacuous(m, enabled, deriv->daughters[i], toks);
		return;
	}
	// this is a lexeme. see if it's vacuous.
	//printf("reconstruct tree: \n"); print_tree(deriv);
	struct dg	*lexd = reconstruct_tree(deriv, NULL);
	assert(lexd);
	struct dg	*rels = dg_path(lexd, lex_rels_path);
	struct dg	*list = dg_hop(rels, lookup_fname("LIST"));
	struct dg	*last = dg_hop(rels, lookup_fname("LAST"));
	if(list != last)return;

	//printf("vacuous word '%s' [%d - %d]\n", deriv->label, deriv->cfrom, deriv->cto);
	struct lexeme	*lex = get_lex_by_name_hash(deriv->label);
	assert(lex);

	if(vacuous_in_scope(lex, deriv, toks))
	{
		mark_token(toks, deriv->cfrom, deriv->cto, 3);
		//printf(" ... in scope.\n");
	}
}

print_tree_with_edge_ids(struct tree	*t, int	indent)
{
	int i;
	for(i=0;i<indent;i++)printf("  ");
	printf("%d	", t->edge_id);
	if(t->cfrom==-1 || t->cto==-1)
		printf("%s    [s=%f]\n", t->label, t->score);
	else printf("%s    [s=%f]   [%d-%d]\n", t->label, t->score, t->cfrom, t->cto);
	/*if(t->ntokens && print_tokens)
	{
		int j;
		for(j=0;j<t->ntokens;j++)
		{
			for(i=0;i<indent;i++)printf("  ");
			printf(" => %s\n", t->tokens[j]);
		}
	}*/
	/*if(t->decoration)
	{
		for(i=0;i<indent;i++)printf("  ");
		print_tree_decoration(t->decoration);
	}*/
	for(i=0;i<t->ndaughters;i++)
		print_tree_with_edge_ids(t->daughters[i], indent+1);
}

int	unmarked_ep(struct mrs_ep	*ep)
{
	if(!strcmp(ep->pred, "udef_q_rel"))return 1;
	else return 0;
}

int is_conj_pred(char	*pred)
{
	if(strstr(pred, "_c_"))return 1;
	if(strstr(pred, "implicit_conj"))return 1;
	return 0;
}

int	quant_init_fc_acceptable(char	*pred, char	*role)
{
	//old version:
	//if(!strstr(pred, "_v_"))return 0;

	// no conjunctions or fragment relations
	if(is_conj_pred(pred))return 0;
	if(strstr(pred, "unknown_rel"))return 0;
	// we're looking for ARG1, ARG2, etc
	if(!strcmp(role, "ARG0"))return 0;
	return 1;
}


/*
int	functor_crawl1(struct mrs	*m, struct mrs_var	*v, int	*enabled)
{
	int i, j, n = 0;
	for(i=0;i<m->nrels;i++)
	{
		struct mrs_ep	*e = &m->rels[i];
		for(j=0;j<e->nargs;j++)
		{
			if(!acceptable_fcpred(e->pred, e->args[j].name))continue;
			struct mrs_var	*arg = e->args[j].value;
			arg = traverse_qeq(m, arg)?:arg;
			if(arg == v && !enabled[i])
			{
				printf("enabled %s by functor-crawl\n", m->rels[i].pred);
				enabled[i] = 1;
				n++;
			}
		}
	}
	return n;
}
*/


int functor_crawl(struct mrs	*m, struct mrs_var	*v, int	*enabled, int	(*acceptable)(char*,char*), int	continue_with_label)
{
	int i, j, N = 0;
	for(i=0;i<m->nrels;i++)
	{
		struct mrs_ep	*ep = &m->rels[i];
		if(is_conj_pred(ep->pred) && !enabled[i])
		{
			struct mrs_var	*arg0 = mrs_find_ep_role(ep, "ARG0");
			struct mrs_var	*lidx = mrs_find_ep_role(ep, "L-INDEX");
			struct mrs_var	*ridx = mrs_find_ep_role(ep, "R-INDEX");
			if(v==lidx || v==ridx)
			{
				enabled[i] = -1;	// -1 means don't try to enable me in the future; verbs taking arg0 as argument will try to enable the whole conjunction, including the out-of-scope conjuncts.
				struct mrs_var	*continue_var = arg0;
				if(continue_with_label)continue_var = ep->lbl;
				if(continue_var != v)
				{
					printf(" ... through %s, also looking for EPs taking %s%d\n", ep->pred, continue_var->type, continue_var->vnum);
					N += functor_crawl(m, continue_var, enabled, acceptable, continue_with_label);
				}
			}
		}
		if(enabled[i]==-1)continue;
		for(j=0;j<ep->nargs;j++)
		{
			if(!acceptable(ep->pred, ep->args[j].name))continue;
			struct mrs_var	*arg = ep->args[j].value;
			arg = traverse_qeq(m, arg)?:arg;
			if(arg == v)
				break;
		}
		if(j==ep->nargs)continue;
		if(!enabled[i])
		{
			printf("enabled %s by functor-crawl\n", ep->pred);
			enabled[i] = 1;
			N++;
		}
	}
	return N;
}

// rough algorithm from delph-in 2013 scope of negation SIG:
//  crawl the graph on the following links:
//    - if X is activated, so is everything sharing X's label
//    - if X is activated, so is everything whose ARG0 is ARGn of X (n > 0)
//    - if X is activated, so is everything whose LBL is (qeq) ARGn of X (n > 0)
//  the initially activated set is the negation cue itself, *plus*:
//     if the negation cue is a quantifier X, activate all *_v_* EPs whose ARGn is X's ARG0
//  after crawling, deactivate the negation cue itself

// to avoid picking up the slashed element of a relative clause, a complication is needed:
//   refuse to pick up an EP that shares its LBL with neg_rel, or a quantifier whose RSTR is qeq neg_rel

struct cue_info
{
	char	*pred;
	int		crawl;
	int		quantifier;
} cue_infos[] = {
	{"never_a_1", 1, 0},
	{"neg_rel", 1, 0},
	{"no_q", 1, 1},
	{"no_a_1", 0, 0},
	{"nor_x_subord", 1, 0},
	{"nothing_n_1_rel", 1, 1},
	{"without_p", 2, 0},
	{"not_x_deg_rel", 3, 1},
	{"and+not_c_rel", 4, 1},
	{"nor_c_rel", 5, 1}
	};
int ncue_infos = sizeof(cue_infos) / sizeof(struct cue_info);

//int alt_find_scope(char	*mrs_file_name, int	from, int	to, struct starsem_token_list	*toks)
int alt_find_scope(struct mrs	*m, struct tree	*deriv, char	*type, int	from, int	to, struct starsem_token_list	*toks)
{
	if(!m)
	{
		fprintf(stderr, "unable to read an MRS; perhaps no parses found?\n");
		return -1;
	}
	assert(m);

	print_mrs(m);

	printf("looking for a negation cue with span < %d : %d >\n", from, to);

	int i, j, ep_neg = -1, ep_neg_cand = -1;
	int should_crawl = 1, crawl_like_quantifier = 0;
	for(i=0;i<m->nrels;i++)
	{
		struct mrs_ep	*ep = &m->rels[i];
		if(ep->cfrom <= from && ep->cto >= to)
		{
			ep_neg_cand = i;
			// don't let gpreds be cues, except neg_rel
			if(ep->pred[0]!='"' && ep->pred[0]!='_' && !strstr(ep->pred, "neg_rel")) continue;
			// always prefer neg_rel if it's available... this helps disambiguate some derivationally decomposed prefixes
			if(strstr(ep->pred, "neg_rel")) { ep_neg = i; continue; }
			if(!strcmp(type, ":form"))
			{
				int k;
				for(k=0;k<ncue_infos;k++)
					if(strstr(ep->pred, cue_infos[k].pred))break;
				if(k==ncue_infos)continue;
				should_crawl = cue_infos[k].crawl;
				crawl_like_quantifier = cue_infos[k].quantifier;
				//if(!strstr(ep->pred, "never_a_1") && !strstr(ep->pred, "neg_rel") && !strstr(ep->pred, "no_q")) continue;
			}
			else if(strstr(ep->pred, "_q_"))continue;
			if(ep_neg == -1)ep_neg = i;
			else printf("conflict: alternate negation cue candidate with given span: %s\n", ep->pred);
		}
	}
	if(ep_neg == -1)
	{
		fflush(stdout);
		fprintf(stderr, "unable to find negation cue EP; bailing out!\n");
		if(ep_neg_cand != -1)fprintf(stderr, "but I did find `%s' where the cue should be\n", m->rels[ep_neg_cand].pred);
		return -1;
	}

	int	enabled[m->nrels];
	bzero(enabled, sizeof(enabled));

	// found the negated predication
	struct mrs_ep	*neg = &m->rels[ep_neg];
	printf("negation cue is: %s\n", neg->pred);

	// tweak starting conditions based on crawl type requested
	switch(should_crawl)
	{
		case	0:
			printf(" ... it was one I shouldn't crawl from.\n");
			return 0;
		case	2:
			// negative preposition;
			// only crawl its ARG2.
			enabled[neg->epnum] = -1;
			struct mrs_var	*arg2 = mrs_find_ep_role(neg, "ARG2");
			mark_by_arg0(m, arg2, enabled, neg);
			break;
		case	3: ;
			// negative degree specifier on a quantifier.
			// plan: find the quantifier (it shares a label with the cue),
			// and treat that as the negation cue.
			struct mrs_var	*lbl = neg->lbl;
			for(i=0;i<m->nrels;i++)
				if(m->rels[i].lbl == lbl && strstr(m->rels[i].pred, "_q_"))
					break;
			if(i==m->nrels)
			{
				fflush(stdout);
				fprintf(stderr, "negative degree specifier, but no quantifier shared its label. :-(\n");
				return -1;
			}
			neg = &m->rels[i];
			crawl_like_quantifier = 1;
			printf("effective negation cue is: %s\n", neg->pred);
			break;
		case	4: ;
			// negative conjunction.
			// plan: find the quantifier of the R-INDEX
			// and treat that as the negation cue.
			struct mrs_var	*r_index = mrs_find_ep_role(neg, "R-INDEX");
			assert(r_index);
			for(i=0;i<m->nrels;i++)
				if(mrs_find_ep_role(&m->rels[i], "ARG0") == r_index && strstr(m->rels[i].pred, "_q_"))
					break;
			if(i==m->nrels)
			{
				fflush(stdout);
				fprintf(stderr, "negative conjunction, but no quantifier for its R-INDEX\n");
				return -1;
			}
			neg = &m->rels[i];
			crawl_like_quantifier = 1;
			printf("effective negation cue is: %s\n", neg->pred);
			break;
		case	5: ; {
			// bi-negative conjunction.
			// plan: mark L-INDEX
			struct mrs_var	*l_index = mrs_find_ep_role(neg, "L-INDEX");
			assert(l_index);
			mark_by_arg0(m, l_index, enabled, neg);
			// find the quantifier of the R-INDEX
			// and treat that as the negation cue.
			struct mrs_var	*r_index = mrs_find_ep_role(neg, "R-INDEX");
			assert(r_index);
			for(i=0;i<m->nrels;i++)
				if(mrs_find_ep_role(&m->rels[i], "ARG0") == r_index && strstr(m->rels[i].pred, "_q_"))
					break;
			if(i==m->nrels)
			{
				fflush(stdout);
				fprintf(stderr, "negative conjunction, but no quantifier for its R-INDEX\n");
				fprintf(stderr, "treating as a simple negative adverb...\n");
			}
			else
			{
				neg = &m->rels[i];
				crawl_like_quantifier = 1;
				printf("effective negation cue is: %s\n", neg->pred);
			}
			} break;
	}

	//  the initially activated set is the negation cue itself, *plus*:
	//     if the negation cue is a quantifier X, activate all suitable EPs whose ARGn is X's ARG0
	if(!enabled[neg->epnum])
		enabled[neg->epnum] = 1;
	if(crawl_like_quantifier)
	{
		// activate all suitable EPs having the negated X as ARGn, n>0
		struct mrs_var	*negated_x = mrs_find_ep_role(neg, "ARG0");
		assert(negated_x != NULL);
		functor_crawl(m, negated_x, enabled, quant_init_fc_acceptable, 0);
	}

	// main loop:
	// apply label and argument crawling until convergence
	int newmarks;
	do
	{
		newmarks = 0;
		for(i=0;i<m->nrels;i++)
		{
			if(enabled[i] != 1)continue;
			struct mrs_ep	*ep = &m->rels[i];
			//    - if X is activated, so is everything sharing X's label
			newmarks += mark_by_label(m, ep->lbl, enabled, neg);
			for(j=0;j<ep->nargs;j++)
			{
				// we're looking for ARG1, ARG2, etc
				if(!strcmp(ep->args[j].name, "ARG0"))
					continue;
				struct mrs_var	*v = ep->args[j].value;
				//    - if X is activated, so is everything whose ARG0 is ARGn of X (n > 0)
				newmarks += mark_by_arg0(m, v, enabled, neg);
				//    - if X is activated, so is everything whose LBL is (qeq) ARGn of X (n > 0)
				newmarks += mark_by_label(m, v, enabled, neg);
				v = traverse_qeq(m, v);
				if(v)newmarks += mark_by_label(m, v, enabled, neg);
			}
			newmarks += functor_crawl(m, ep->lbl, enabled, acceptable_fcpred, 1);
			newmarks += functor_crawl(m, mrs_find_ep_role(ep, "ARG0"), enabled, acceptable_fcpred, 1);
		}
		//printf("%d new marks\n", newmarks);
	} while(newmarks);

	// disable zero pronouns implied by imperatives (in particular "don't")
	disable_zero_pronouns(m, enabled);

	for(i=0;i<m->nrels;i++)
		if(enabled[i]==1 && !unmarked_ep(&m->rels[i])) mark_token(toks, m->rels[i].cfrom, m->rels[i].cto, 3);

	// vacuous word processing
	// three times so that stacks of auxilliaries can all get marked
	deriv = depunct_tree(deriv, NULL);
	decorate_deriv(deriv, toks);
	//print_tree_with_edge_ids(deriv, 0);
	mark_vacuous(m, enabled, deriv, toks);
	decorate_deriv(deriv, toks);
	mark_vacuous(m, enabled, deriv, toks);
	decorate_deriv(deriv, toks);
	mark_vacuous(m, enabled, deriv, toks);
}
