#ifndef	_DAG_H
#define	_DAG_H

#include	<stdio.h>
#include	<assert.h>
#include	"type.h"

//#define COPY_COUNTS
//#define COPY_PATHS
#define	UNIFY_PATHS

// tomabechi dgs

#define	MAX_GEN	0xffffffff
#define	MAX_GEN_RESET 4000000000UL

//#define	MAX_GEN	0xffff
//#define	MAX_GEN_RESET 55000

struct darc
{
	int			label;
	struct dg	*dest;
};

// ideas for reducing size of DAG node structure:
// 1. represent types by 32-bit or even 16-bit??? numbers
//		pros: save a lot of space. access to type-internal fields is rare. glb hash lookup can be done without dereferencing.
//		cons: access to type-internal fields is slow (indirecting via a big array of pointers is likely to cause at least one cache miss). temporary string types may require a little thought.
// 2. temporary slots are somewhat redundant
//      e.g., when a dag is forwarded, we don't need space to store type, carcs, or copy.
//			(but: subsume() uses both copy and forward as pointers. does it need to? I think so...?)
//		we could likely get away with two pointer-value scratch slots (16 bytes) instead of the current 4.
//      we frequently need both 'type' and 'carcs'
//		when copying a temporary structure, we use 'copy' also.
//			... but once 'copy' gets set, we could perhaps overwrite 'type' and 'carcs'... if we don't mind this being a destructive operation.  most use-cases only want to copy out of a temporary dag once... transfer and lattice mapping may want to copy several outputs, but we could copy the whole rule and then extract the relevant parts if this were a big win.
// 3. how about removing temporary storage from the dag node almost altogether? e.g. have a 'struct dgtemp' that has the relevant fields, and just a pointer to it in 'struct dg'.
//		pros: significant overall space reduction.
//		cons: how do we implement permanent forwarding? accessing temp fields may be slow (although if the working set of (reusable) temp structures is small, it may stay in cache)
//		one step further: reference from 'struct dg' to 'struct dgtmp' could be by any number of mechanisms... e.g. an int, or a (potentially thread-local) hash table, or?

// 48 bytes... we expect to have roughly one of these per dg node being actively manipulated this generation.
// for a typical unification between 2 dags which each have 100 nodes, that's around 200 structures = 24k bytes.
// we can't really hope that it will fit in L1 all the time, but it certainly should be in L2 near-constantly.

struct dg_tmp_fields
{
	struct dg	*forward;
	struct dg	*copy;
	struct type	*newtype;
	struct darc	*carcs;
	unsigned int	gen;
	unsigned char	ncarcs;
} __attribute__((packed));

struct dg_mrstmp
{
	struct mrs_var	*var;
} __attribute__((packed));

struct dg_tmp
{
	struct dg	*owner;
	struct dg_tmp_fields	tmp;
	unsigned char	pad[3];
} __attribute__((packed));

// 8 bytes!
struct tinydg
{
	unsigned int	type:28;
	unsigned int	tmp:12;	// we expect the number of active 'tmp' structures to be less than 1000
	unsigned int	narcs:8;
#define	TINY_DAG_BUILTIN_ARCS	1
	unsigned short	label[TINY_DAG_BUILTIN_ARCS];
#define	TINY_DAG_EXTRA_LABELS_SIZE_FOR_COUNT(n)	( (((n)-TINY_DAG_BUILTIN_ARCS) * 2 + 3) / 4 * 4)
} __attribute__((packed));

// 16 btes
struct smalldg
{
	struct type	*xtype;
	unsigned short	tmp;	// we expect the number of active 'tmp' structures to be less than 1000
	unsigned short	narcs;
#define	SMALL_DAG_BUILTIN_ARCS	2
	unsigned short	label[SMALL_DAG_BUILTIN_ARCS];
#define	SMALL_DAG_EXTRA_LABELS_SIZE_FOR_COUNT(n)	( (((n)-SMALL_DAG_BUILTIN_ARCS) * 2 + 3) / 4 * 4)
} __attribute__((packed));

/*
struct dg
{
	struct type	*type;
	struct type	*xtype;
	struct dg	*forward;
	struct darc	*carcs;
	struct dg	*copy;
	unsigned int	gen;
	unsigned char	narcs, ncarcs;
	//unsigned char	label[2];
	//unsigned short	label[1];
#define	BUILTIN_ARCS	1
	unsigned short	label[BUILTIN_ARCS];
#define	DAG_EXTRA_LABELS_SIZE_FOR_COUNT(n)	((((n)-1)*2+3)/4 * 4)
//#define	DAG_EXTRA_LABELS_SIZE_FOR_COUNT(n)	(((n)>BUILTIN_ARCS)?( (((n)-BUILTIN_ARCS) * 2 + 3) / 4 * 4):(0))
} __attribute__((packed));

#define	DARCS(dg1)	((struct dg**)((void*)(dg1)+sizeof(struct dg) + DAG_EXTRA_LABELS_SIZE(dg1)))
#define	DAG_EXTRA_LABELS_SIZE(dg1)	DAG_EXTRA_LABELS_SIZE_FOR_COUNT((dg1)->narcs)
*/

struct dg
{
	struct type	*xtype;
	struct dg_tmp_fields	tmp;
	unsigned int	narcs:7, is_perma_forwarded:1;
#define	BUILTIN_ARCS	1
	unsigned short	label[BUILTIN_ARCS];
#define	DAG_EXTRA_LABELS_SIZE_FOR_COUNT(n)	((((n)-1)*2+3)/4 * 4)
//#define	DAG_EXTRA_LABELS_SIZE_FOR_COUNT(n)	(((n)>BUILTIN_ARCS)?( (((n)-BUILTIN_ARCS) * 2 + 3) / 4 * 4):(0))
} __attribute__((packed));

#define	DAG_EXTRA_LABELS_SIZE(dg1)	DAG_EXTRA_LABELS_SIZE_FOR_COUNT((dg1)->narcs)

// accessor macros
#define	DARCS(dg1)	((struct dg**)((void*)(dg1)+sizeof(struct dg) + DAG_EXTRA_LABELS_SIZE(dg1)))
#define	GET_ORIG_TYPE(dg)	((dg)->xtype)
#define	HAS_TMP(dg)		((dg)->tmp.gen == generation)
static inline struct dg_tmp_fields	*GET_VALID_TMP(struct dg	*d)
{
	extern unsigned int generation;
	if(!HAS_TMP(d))
	{
		struct dg_tmp_fields	*tmp = &d->tmp;
		//bzero(tmp, sizeof(*tmp));
		tmp->copy = tmp->forward = NULL;
		tmp->carcs = NULL;
		tmp->ncarcs = 0;
		tmp->newtype = d->xtype;
		tmp->gen = generation;
	}
	return &d->tmp;
}
#define	INVALIDATE_TMP(dg)	(dg)->tmp.gen = 0

#define HAS_MRSTMP(dg)	((dg)->tmp.gen == 1)
static inline struct dg_mrstmp	*GET_VALID_MRSTMP(struct dg	*d)
{
	if(!HAS_MRSTMP(d))
	{
		d->tmp.gen = 1;
		struct dg_mrstmp	*mt = (struct dg_mrstmp*)&d->tmp;
		mt->var = NULL;
	}
	return (struct dg_mrstmp*)&d->tmp;
}

static inline void SET_XTYPE(struct dg	*dg, struct type	*ty)
{
	dg->tmp.gen = 0;
	dg->xtype = ty;
}

// basic dag operations
int		unify_dg_infrom(struct dg	*dg1, struct dg	*dg2);
struct dg	*unify_dg(struct dg	*a, struct dg	*b, int	arg);
struct dg	*unify_dg_noss(struct dg	*dg1, struct dg	*dg2, int	arg);
struct dg	*atomic_dg(char	*name);
struct dg	*copy_dg(struct dg	*d);
struct dg	*trim_copy(struct dg	*d);
int		lookup_fname(char	*label);
void		null_generation();
void		print_dg(struct dg	*d);
void		unify_backtrace();
void		introduce_feature(int	label, int	type);
struct dg	*add_dg_hop(struct dg	*d, int	label, struct dg	*value);
struct dg	*add_dg_feature(struct dg	*d, char	*label, struct dg	*value);
struct dg	*add_dg_path(struct dg	*d, int	*path, int	len, struct dg	*target);
int		init_fnames();
int		equivalent_dg(struct dg	*a, struct dg	*b, int Fw, int Bk);
void		setup_extra_type_dgs();
void		setup_carcs_stack();
struct dg	*wellform_dg(struct dg	*dg);
struct dg	*walk_dg(struct dg	*dg, int *save, ...);
void		output_lui_dg(struct dg	*dg, char *wtitle);

struct dg	*generalize(struct dg	*a, struct dg	*b);

// internal operations
int			unify1(struct dg	*dg1, struct dg	*dg2);
struct dg	*copy_dg_with_comp_arcs(struct dg	*dg);

// for hyperactive parsing
int		unify_dg_tmp(struct dg	*dg1, struct dg	*dg2, int	arg);
int		unify_dg_defaults(struct dg	*indef, struct dg	*def);
struct dg	*finalize_tmp(struct dg	*d, int trim);
struct dg	*finalize_tmp_noss(struct dg	*d, int trim);
int		check_cycles(struct dg	*dg);
void		forget_tmp();

// quickcheck operations
#include	"qc.h"

// slab operations
void	clear_slab();
void	commit_slab();
void	*slab_realloc(void	*ptr, int	oldlen, int	newlen);
void	print_slab_stats();

struct slab_state
{
	int	noldslabs;
	void	*slab, *last;
	int	slabu;
};

static inline void	get_slabstate(struct slab_state	*s)
{
	extern void	*slab, *last;
	extern int	slabu, noldslabs;
	s->noldslabs = noldslabs;
	s->slab = slab;
	s->last = last;
	s->slabu = slabu;
}

static inline void	set_slabstate(struct slab_state	*s)
{
	extern void	*slab, *last;
	extern int	slabu, noldslabs;
	noldslabs = s->noldslabs;
	slab = s->slab;
	last = s->last;
	slabu = s->slabu;
}

static inline	void	*slab_alloc(int	len) __attribute__ ((malloc));

static inline	void	*slab_alloc(int	len)
{
	extern void	*last, *slab;
	extern int	slabu, currslabsize;
	extern void	next_slab();

	assert(len>=0);
	assert((len%4)==0);

	// align these special cases
	if(len==48)
	{
		if((unsigned long)last % 48)
		{
			slabu += 48 - ((unsigned long)last % 48);
			last += 48 - ((unsigned long)last % 48);
		}
	}
	/*else if(!(len&7))
	{
		if((unsigned int)last & 7)
		{
			slabu += 8 - ((unsigned int)last & 7);
			last += 8 - ((unsigned int)last & 7);
		}
	}*/

	if(len>currslabsize)
	{
		fprintf(stderr, "allocation request %d too big for slab\n", len);
		exit(-1);
	}
	if(slabu+len > currslabsize)next_slab();
	last = slab + slabu;
	slabu += len;
	return last;
}

static inline struct dg	*p_dereference_dg(struct dg	*dg)
{
	while(dg && dg->is_perma_forwarded)dg = (struct dg*)dg->xtype;
	return dg;
}

static inline struct dg	*dereference_dg(struct dg	*dg)
{
	struct dg	*fw = dg;
	extern unsigned int	generation;

	if(fw && !fw->is_perma_forwarded && (!HAS_TMP(fw) || !GET_VALID_TMP(fw)->forward))return fw;

	while(1)
	{
		// step 'fw' twice
		if(fw->is_perma_forwarded)fw = (struct dg*)fw->xtype;
		else if(HAS_TMP(fw)) { struct dg_tmp_fields	*tmp = GET_VALID_TMP(fw); if(tmp->forward)fw = tmp->forward; else return fw; }
		if(fw->is_perma_forwarded)fw = (struct dg*)fw->xtype;
		else if(HAS_TMP(fw)) { struct dg_tmp_fields	*tmp = GET_VALID_TMP(fw); if(tmp->forward)fw = tmp->forward; else return fw; }
		else return fw;
		
		// step 'dg' once
		if(dg->is_perma_forwarded)dg = (struct dg*)dg->xtype;
		else dg = GET_VALID_TMP(dg)->forward;
		if(dg == fw)
		{
			fprintf(stderr, "XXX deref cycle\n");
			return 0;
		}
	}
}

static inline struct dg	*dg_hop(struct dg	*d, int	label)
{
	int	i, j, k, l;
	extern unsigned int	generation;
	struct dg	**arcs;

	if(!d)return NULL;
	d = dereference_dg(d);
	arcs = DARCS(d);

	for(i=0,j=d->narcs;i<j;)
	{
		k = (i+j)>>1;
		//printf("guess %d: %d\n", k, arcs[k].label);
		l = d->label[k];
		if(l==label){d=arcs[k];goto ret;}
		else if(l>label)j = k;
		else i = k+1;
	}
	if(HAS_TMP(d))
	{
		struct dg_tmp_fields	*tmp = GET_VALID_TMP(d);
		for(i=0,j=tmp->ncarcs;i<j;)
		{
			k = (i+j)>>1;
			l = tmp->carcs[k].label;
			if(l==label){d=tmp->carcs[k].dest;goto ret;}
			else if(l>label)j = k;
			else i = k+1;
		}
	}
	return NULL;
ret:
	return dereference_dg(d);
}

#endif
