load_reading(struct s_exp	*s_reading, struct mrs	*m, int	*is_hole, int	*ep_for_lbl, int	*var_for_num)
{
	int j;
	for(j=0;j<s_reading->len;j++)
	{
		struct s_exp	*s_plugging = s_reading->list[j];
		assert(s_plugging->type == s_exp_list && s_plugging->len == 3);
		struct s_exp	*s_b = s_plugging->list[1], *s_c = s_plugging->list[2];
		assert(s_b->type == s_exp_id && s_c->type == s_exp_id);
		int	b = atoi(s_b->value), c = atoi(s_c->value);
		assert(b >= 0 && b <= m->vlist.nvars);
		assert(c >= 0 && c <= m->vlist.nvars);
		b = var_for_num[b];
		c = var_for_num[c];
		assert(b != -1 && c != -1);
		assert(is_hole[b]);
		int	epc = ep_for_lbl[c];
		assert(epc != -1);
		//printf("h%d filled by h%d = ep%d = %s\n", b, c, epc, m->rels[epc].pred);
		ep_for_lbl[b] = epc;
	}
}

void	find_span(struct mrs	*m, int	*ep_for_lbl, int	epi, int	inclusive, int	*Left, int	*Right, struct starsem_token_list	*toks)
{
	struct mrs_var	*lbl = m->rels[epi].lbl;
	int i, j;
	for(j=0;j<m->nrels;j++)
	{
		struct mrs_ep	*e = m->rels+j;
		if(e->lbl != lbl)continue;
		if(inclusive && e->cfrom < *Left)*Left = e->cfrom;
		if(inclusive && e->cto > *Right)*Right = e->cto;
		mark_token(toks, e->cfrom, e->cto, 2);
		for(i=0;i<e->nargs;i++)
		{
			struct mrs_var	*arg = e->args[i].value;
			if(ep_for_lbl[arg->vnum] == -1)continue;
			find_span(m, ep_for_lbl, ep_for_lbl[arg->vnum], 1, Left, Right, toks);
		}
	}
}

int	count_outscoped(struct mrs	*m, int	*ep_for_lbl, int	epi)
{
	struct mrs_var	*lbl = m->rels[epi].lbl;
	int i, j, n = 0;
	for(j=0;j<m->nrels;j++)
	{
		struct mrs_ep	*e = m->rels+j;
		if(e->lbl != lbl)continue;
		n += 1;
		for(i=0;i<e->nargs;i++)
		{
			struct mrs_var	*arg = e->args[i].value;
			if(ep_for_lbl[arg->vnum] == -1)continue;
			n += count_outscoped(m, ep_for_lbl, ep_for_lbl[arg->vnum]);
		}
	}
	return n;
}

int	outscopes_via_verb(struct mrs	*m, int	*ep_for_lbl, int	hi_ep, int	lo_ep, int	via_verb)
{
	if(hi_ep == lo_ep)return via_verb?0:1;
	struct mrs_var	*hi_lbl = m->rels[hi_ep].lbl;
	int i, j;
	//printf(" ... does %s outscope %s? [via_verb = %d]\n", m->rels[hi_ep].pred, m->rels[lo_ep].pred, via_verb);
	for(j=0;j<m->nrels;j++)
	{
		struct mrs_ep	*e = m->rels+j;
		if(e->lbl != hi_lbl)continue;
		int	is_verb = strstr(e->pred, "_v_")?1:0;
		int still_need_verb = via_verb && !is_verb;
		// look at each scopal argument of each EP with the hi label
		for(i=0;i<e->nargs;i++)
		{
			struct mrs_var	*arg = e->args[i].value;
			if(ep_for_lbl[arg->vnum] == -1)continue;
			if(outscopes_via_verb(m, ep_for_lbl, ep_for_lbl[arg->vnum], lo_ep, still_need_verb))
				return 1;
		}
	}
	return 0;
}

char	*mrs_find_var_prop(struct mrs_var	*v, char	*prop)
{
	int i;
	for(i=0;i<v->nprops;i++)
		if(!strcasecmp(v->props[i].name, prop))
			return v->props[i].value;
	return NULL;
}

int	eligible(struct mrs	*m, int	*ep_for_lbl, int	neg_epi)
{
	/* 2. When the negation cue is NP negation, do not choose any scope in
	which the negation cue scopes higher than verbs which outscope the
	finite verb that the negated NP is an argument of.  */

	// is this NP negation?
	struct mrs_ep	*noq = &m->rels[neg_epi];
	if(!strstr(noq->pred, "no_q"))return 1;
	struct mrs_var	*q = noq->lbl;
	struct mrs_var	*x = mrs_find_ep_role(noq, "ARG0");

	// look at each finite verb that the negated NP is an argument of
	int i, j;
	for(i=0;i<m->nrels;i++)
	{
		struct mrs_ep	*e = &m->rels[i];
		if(!strstr(e->pred, "_v_"))continue;
		for(j=0;j<e->nargs;j++)if(e->args[j].value == x)break;
		if(j==e->nargs)continue;
		// `e' is a verb that the negated NP is an argument of
		// is it finite?
		struct mrs_var	*ev = mrs_find_ep_role(e, "ARG0");
		char	*tense = mrs_find_var_prop(ev, "TENSE");
		assert(tense != NULL);
		if(!strcasecmp(tense, "untensed"))continue;	// no
		// yes.
		//printf("negated X is argument of '%s'; does negator outscope it via a verb?\n", e->pred);
		//if(outscopes_via_verb(m, ep_for_lbl, q->vnum, e->lbl->vnum, 1))
		if(outscopes_via_verb(m, ep_for_lbl, neg_epi, i, 1))
		{
			//printf("yes.\n");
			return 0;
		}
		//printf("no.\n");
	}
	return 1;
}

print_scoped_mrs(struct mrs	*m1, int	*ep_for_lbl)
{
	struct mrs	*m = copy_mrs(m1);
	m->nhcons = 0;
	int i,j;
	int map(int vnum) { if(ep_for_lbl[vnum] == -1)return vnum; else return m->rels[ep_for_lbl[vnum]].lbl->vnum; }
	struct mrs_var	*vmap(struct mrs_var	*vin) { return m->vlist.vars[map(vin->vnum)]; }
	m->ltop = vmap(m->ltop);
	for(i=0;i<m->nrels;i++)
	{
		struct mrs_ep	*e = &m->rels[i];
		for(j=0;j<e->nargs;j++)
			e->args[j].value = vmap(e->args[j].value);
	}
	print_mrs(m);
}

int find_scope(char	*mrs_file_name, char	*scope_file_name, int	from, int	to, struct starsem_token_list	*toks)
{
	fflush(stdout);
	clear_slab();
	clear_mrs();
	FILE	*mf = fopen(mrs_file_name, "r");
	if(!mf) { perror(mrs_file_name);return -1; }
	struct mrs	*m = read_mrs(mf);
	fclose(mf);
	if(!m)
	{
		fprintf(stderr, "unable to read an MRS; perhaps no parses found?\n");
		return -1;
	}
	assert(m);
	FILE	*pf = fopen(scope_file_name, "r");
	if(!pf) { perror(scope_file_name);return -1; }
	fseek(pf, 0, SEEK_END);
	size_t	l = ftell(pf);
	rewind(pf);
	char	*buf = malloc(l+2);
	buf[l] = ')';buf[l+1] = 0;	// add a close paren, in case we lost it by `head'-ing the utool output
	assert(l == fread(buf, 1, l, pf));
	fclose(pf);

extern char mrs_line_breaks;
	mrs_line_breaks = '\n';
	//print_mrs(m);

	//printf("looking for a negation cue with span < %d : %d >\n", from, to);

	int i, ep_neg = -1, ep_neg_cand = -1;
	for(i=0;i<m->nrels;i++)
	{
		struct mrs_ep	*ep = &m->rels[i];
		if(ep->cfrom <= from && ep->cto >= to)
		{
			ep_neg_cand = i;
			if(!strstr(ep->pred, "never_a_1") && !strstr(ep->pred, "neg_rel") && !strstr(ep->pred, "no_q"))continue;
			if(ep_neg == -1)ep_neg = i;
			else printf("conflict: alternate negation cue candidate with given span: %s\n", ep->pred);
		}
	}
	if(ep_neg == -1)
	{
		fflush(stdout);
		fprintf(stderr, "unable to find negation cue EP; bailing out!\n");
		if(ep_neg_cand != -1)fprintf(stderr, "but I did find `%s' where the cue should be\n", m->rels[ep_neg_cand].pred);
		return -1;
	}
	printf("negation cue is: %s\n", m->rels[ep_neg].pred);

	struct mrs_ep	*neg = m->rels+ep_neg;
	int nscopal_args = 0;
	for(i=0;i<neg->nargs;i++)
		if(neg->args[i].value->is_const == 0 && !strcasecmp(neg->args[i].value->type, "h"))
			nscopal_args++;
	if(nscopal_args == 0)
	{
		fflush(stdout);
		fprintf(stderr, "error: negation cue is not scope-bearing in this analysis!\n");
		return -1;
	}

	int	ep_for_lbl[m->vlist.nvars];
	for(i=0;i<m->vlist.nvars;i++)ep_for_lbl[i] = -1;
	for(i=0;i<m->nrels;i++)
		ep_for_lbl[m->rels[i].lbl->vnum] = i;

	int	is_hole[m->vlist.nvars];
	for(i=0;i<m->vlist.nvars;i++)
		if(ep_for_lbl[i] != -1)is_hole[i] = 0;
		else if(m->vlist.vars[i]->is_const)is_hole[i] = 0;
		else if(m->vlist.vars[i]->name[0] != 'h')is_hole[i] = 0;
		else is_hole[i] = 1;

	int	var_for_num[m->vlist.nvars + 1];
	for(i=0;i<=m->vlist.nvars;i++)var_for_num[i] = -1;
	for(i=0;i<m->vlist.nvars;i++)
		if(!m->vlist.vars[i]->is_const && m->vlist.vars[i]->name[0]=='h' && isdigit(m->vlist.vars[i]->name[1]))
		{
			int	v = atoi(m->vlist.vars[i]->name+1);
			assert(v>=0 && v<=m->vlist.nvars);
			var_for_num[v] = i;
		}

	char	*bcopy = buf;
	struct s_exp	*s = parse_lisp(&bcopy, 1);
	assert(s != NULL);
	assert(s->type == s_exp_list);

	int	max_outscoped = -1, best_reading = -1, nbest = 0;
	int	*noutscoped = malloc(sizeof(int) * s->len);
	assert(s->len > 0);
	printf("considering %d scopings\n", s->len);
	for(i=0;i<s->len;i++)
	{
		struct s_exp	*s_reading = s->list[i];
		assert(s_reading->type == s_exp_list);
		load_reading(s_reading, m, is_hole, ep_for_lbl, var_for_num);
		if(!eligible(m, ep_for_lbl, ep_neg))
		{
			noutscoped[i] = -1;
			continue;
		}
		int	outscoped = count_outscoped(m, ep_for_lbl, ep_neg);
		noutscoped[i] = outscoped;
		//printf("reading %d: negation cue outscopes %d EPs\n", i, outscoped);
		if(max_outscoped < outscoped)
		{
			max_outscoped = outscoped;
			best_reading = i;
			nbest = 1;
		}
		else if(max_outscoped == outscoped)nbest++;
	}

	assert(nbest > 0);
	if(nbest != 1)
	{
		printf("multiple maximally high scopings for the negation cue (%d readings where it outscopes %d EPs)\n", nbest, max_outscoped);
	}
	printf("scoping %d has negation highest\n", best_reading);

	int nprinted = 0;
	for(i=0;i<s->len;i++)
	{
		if(noutscoped[i] < max_outscoped)continue;
		if(nprinted >= 10)
		{
			printf("... not showing %d additional tied-for-best scopings\n", nbest - nprinted);
			break;
		}
		nprinted++;
		struct s_exp	*s_reading = s->list[i];
		assert(s_reading->type == s_exp_list);
		load_reading(s_reading, m, is_hole, ep_for_lbl, var_for_num);
		print_scoped_mrs(m, ep_for_lbl);
		int	left = 100000, right = -1;
		find_span(m, ep_for_lbl, ep_neg, 1, &left, &right, toks);
		printf("scope of negation for this scoping: %d - %d\n", left, right);
	}
}

struct mrs_ep	*mrs_find_ep_by_lbl(struct mrs	*m, struct mrs_var	*lbl)
{
	int i;
	for(i=0;i<m->nhcons;i++)
		if(m->hcons[i].type == hcons_qeq && m->hcons[i].hi == lbl)
			lbl = m->hcons[i].lo;
	for(i=0;i<m->nrels;i++)
		if(m->rels[i].lbl == lbl)return &m->rels[i];
	return NULL;
}

struct mrs_ep	*mrs_find_ep_by_arg0(struct mrs	*m, struct mrs_var	*x)
{
	int i;
	for(i=0;i<m->nrels;i++)
		if(mrs_find_ep_role(&m->rels[i], "ARG0") == x)
		{
			if(strstr(m->rels[i].pred, "_q_"))continue;
			return &m->rels[i];
		}
	return NULL;
}

find_span_of_pred_with_args(struct mrs	*m, struct mrs_ep	*e, struct starsem_token_list	*toks)
{
	int i;
	printf("in scope: %s    %d - %d\n", e->pred, e->cfrom, e->cto);
	mark_token(toks, e->cfrom, e->cto, 3);
	for(i=0;i<e->nargs;i++)
	{
		if(!strcmp(e->args[i].name, "ARG0"))continue;
		struct mrs_var	*x = e->args[i].value;
		if(!strcmp(x->type, "x") || !strcmp(x->type, "e"))
		{
			struct mrs_ep	*arg_ep = mrs_find_ep_by_arg0(m, x);
			if(!arg_ep)continue;
			//assert(arg_ep != NULL);
			find_span_of_pred_with_args(m, arg_ep, toks);
		}
		else if(!strcmp(x->type, "h"))
		{
			struct mrs_ep	*arg_ep = mrs_find_ep_by_lbl(m, x);
			if(!arg_ep)
			{
				fprintf(stderr, "hole argument %s%d had no obvious filler\n", x->type, x->vnum);
				continue;
			}
			//assert(arg_ep != NULL);
			find_span_of_pred_with_args(m, arg_ep, toks);
		}
	}
}
